@@ -1,8 +1,19 @@
-__version__ = "1.0.1.3"
+__version__ = "1.1.0.0"
 __author__ = "Jon Kristian Dahl"
 __credits__ = "Noritaka Shimizu, Jørgen Eriksson Midtbø"
 
+"""
+Version legend:
+a.b.c.d
+
+a: major release
+b: new functionality added
+c: new feature to existing functionality
+d: bug fixes
+"""
+
 from .kshell_utilities import *
 from .general_utilities import *
 from .kshell_exceptions import *
+from .count_dim import *
 # from .low_energy_enhancement import * # Not ready for ver. 1.0
\ No newline at end of file
