@@ -1,4 +1,4 @@
-import sys
+import sys, time
 from typing import Union, Tuple
 from fractions import Fraction
 import numpy as np
@@ -9,30 +9,37 @@ def create_spin_parity_list(
     parities: np.ndarray
     ) -> list:
     """
-    Example list:
-    [[1, +1], [3, +1], [5, +1], [7, +1], [9, +1], [11, +1], [13, +1]].
+    Pair up input spins and parities in a list of lists.
 
     Parameters
     ----------
-    spins:
+    spins : np.ndarray
         Array of spins for each energy level.
 
-    parities:
+    parities : np.ndarray
         Array of corresponding parities for each energy level.
 
     Returns
     -------
-    spins_parities:
+    spins_parities : list
         A nested list of spins and parities [[spin, parity], ...] sorted
         with respect to the spin. N is the number of unique spins in
         'spins'.
-    """
-    unique_spins, unique_spins_idx = np.unique(spins, return_index=True)
-    spins_parities = np.empty((len(unique_spins), 2))
-    spins_parities[:, 0] = unique_spins
-    spins_parities[:, 1] = parities[unique_spins_idx]
 
-    return spins_parities.tolist()  # Convert to list since list.index is much faster than np.where.
+    Examples
+    --------
+    Example list:
+    ``` python
+    [[1, +1], [3, +1], [5, +1], [7, +1], [9, +1], [11, +1], [13, +1]]
+    ```
+    """
+    spin_parity_list = []
+    for i in range(len(spins)):
+        if (tmp := [int(spins[i]), int(parities[i])]) in spin_parity_list:
+            continue
+        spin_parity_list.append(tmp)
+
+    return spin_parity_list
 
 def div0(numerator, denominator):
     """
@@ -56,88 +63,87 @@ def gamma_strength_function_average(
     Ex_min: Union[float, int],
     Ex_max: Union[float, int],
     multipole_type: str = "M1",
-    initial_or_final: str = "initial"
+    initial_or_final: str = "initial",
+    plot: bool = False
     ) -> Tuple[np.ndarray, np.ndarray]:
     """
+    Calculate the gamma strength function averaged over spins and
+    parities.
+    
     Author: Jørgen Midtbø.
-    Modified by: Jon Dahl.
-    Notes from Jørgen:
-        20171009: Updated the way we average over Ex, J, parity_initial to only count pixels with non-zero gSF.
-        20170815: This function returns the strength function the way we now think is the correct way:
-        By taking only the partial level density corresponding to the specific (Ex, J, parity_initial) pixel in the
-        calculation of the strength function, and then averaging over all three variables to produce
-        <f(Eg)>.
-        This code was first developed in the script strength_function_individual_Jpi.py
+    Modified by: GaffaSnobb.
 
-        # # Update 20170915: Realized a problem with summing vs averaging, adding this list to fix that
-        # # Update 20170920: Realized that the fix was wrong, the original was correct.
-        # Ex_already_seen = []
-        # for i_Ex in range(Nbins):
-        #     Ex_already_seen.append([])
-
-        # 20170920: We thought this was more correct, but now think not.
-        # if not Ex in Ex_already_seen[i_Eg]:
-        #     B_pixel_count[i_Ex,i_Eg,spin_parity_idx] += 1
-        #     Ex_already_seen[i_Eg].append(Ex)
-
-    TODO: Ex_final or Ex_initial? Ask about this!
+    TODO: Ex_final or Ex_initial?
     TODO: Figure out the pre-factors.
 
     Parameters
     ----------
-    levels:
+    levels : np.ndarray
         Array containing energy, spin, and parity for each excited
         state. [[E, 2*spin, parity], ...].
 
-    transitions:
+    transitions : np.ndarray
+        OLD:
         Mx8 array containing [2*spin_final, parity_initial, Ex_final,
         2*spin_initial, parity_initial, Ex_initial, E_gamma, B(.., i->f)]
 
-    bin_width:
+        NEW:
+        [2*spin_initial, parity_initial, Ex_initial, 2*spin_final,
+        parity_final, Ex_final, E_gamma, B(.., i->f), B(.., f<-i)]
+
+    bin_width : Union[float, int]
         The width of the energy bins. A bin width of 0.2 contains 20
         states of uniform spacing of 0.01.
 
-    Ex_min:
+    Ex_min : Union[float, int]
         Lower limit for initial level excitation energy, usually in MeV.
 
-    Ex_max:
+    Ex_max : Union[float, int]
         Upper limit for initial level excitation energy, usually in MeV.
 
-    multipole_type:
+    multipole_type : str
         Choose whether to calculate for 'E1', 'M1' or 'E2'. NOTE:
         Currently only M1 is implemented.
 
-    initial_or_final:
+    initial_or_final : str
         Choose whether to use the energy of the initial or final state
         for the transition calculations. NOTE: This will be removed in
         a future release since the correct alternative is to use the
         initial energy.
 
+    plot : bool
+        Toogle plotting on / off.
+
     Variables
     ---------
-    Ex:
+    Ex : np.ndarray 
         The excitation energy of all levels.
 
-    Ex_final:
-        The excitation energy of the final state of a transition.
-
-    Ex_initial:
+    Ex_initial : np.ndarray
         The excitation energy of the initial state of a transition.
 
-    spins:
+    spins : np.ndarray
         The spins of all levels.
 
-    parities:
+    parities : np.ndarray
         The parities of all levels.
 
     Returns
     -------
-    bins:
+    bins : np.ndarray
         The bins corresponding to gSF_ExJpiavg (x values for plot).
         
-    gSF_ExJpiavg:
+    gSF_ExJpiavg : np.ndarray
         The gamma strength function.
     """
+    if (Ex_min < 0) or (Ex_max < 0):
+        msg = "Ex_min and Ex_max cannot be negative!"
+        raise ValueError(msg)
+
+    if Ex_max < Ex_min:
+        msg = "Ex_max cannot be smaller than Ex_min!"
+        raise ValueError(msg)
+
     prefactors = {   # Factor from the def. of the GSF.
         "M1": 11.5473e-9, # [1/(mu_N**2*MeV**2)].
         "E1": 1.047e-6
@@ -151,9 +157,16 @@ def gamma_strength_function_average(
     Ex, spins, parities = np.copy(levels[:, 0]), levels[:, 1], levels[:, 2]
     
     if initial_or_final == "initial":
-        Ex_initial_or_final = np.copy(transitions[:, 5])   # To avoid altering the raw data.
-    elif initial_or_final == "final":
         Ex_initial_or_final = np.copy(transitions[:, 2])   # To avoid altering the raw data.
+        spin_initial_or_final_idx = 0
+        parity_initial_or_final_idx = 1
+    elif initial_or_final == "final":
+        """
+        NOTE: This option will be removed in a future release.
+        """
+        Ex_initial_or_final = np.copy(transitions[:, 5])   # To avoid altering the raw data.
+        spin_initial_or_final_idx = 3
+        parity_initial_or_final_idx = 4
     else:
         msg = "'initial_or_final' must be either 'initial' or 'final'."
         msg += f" Got {initial_or_final}"
@@ -193,10 +206,10 @@ def gamma_strength_function_average(
     bin_array = np.linspace(0, bin_width*n_bins, n_bins + 1) # Array of lower bin edge energy values
     bin_array_middle = (bin_array[0: -1] + bin_array[1:])/2 # Array of middle bin values
     """
-    Ex_min_idx = int(np.floor(Ex_min/bin_width)) 
-    Ex_max_idx = int(np.floor(Ex_max/bin_width))    
+    Ex_min_idx = int(Ex_min/bin_width)
+    Ex_max_idx = int(Ex_max/bin_width)
     n_bins = int(np.ceil(Ex_max/bin_width)) # Make sure the number of bins cover the whole Ex region.
-    Ex_max = bin_width*n_bins # Adjust Ex_max to match the round-off in the bin width. NOTE: Unsure if this is needed.
+    # Ex_max = bin_width*n_bins # Adjust Ex_max to match the round-off in the bin width. NOTE: Unsure if this is needed.
 
     """
     B_pixel_sum[Ex_final_idx, E_gamma_idx, spin_parity_idx] contains the
@@ -205,12 +218,12 @@ def gamma_strength_function_average(
     spin_parity_idx bin. B_pixel_counts counts the number of transitions
     within the same bins.
     """
-    spin_parity_list = create_spin_parity_list(spins, parities)
+    spin_parity_list = create_spin_parity_list(spins, parities) # To create a unique index for every [spin, parity] pair.
     n_unique_spin_parity_pairs = len(spin_parity_list)
     B_pixel_sum = np.zeros((n_bins, n_bins, n_unique_spin_parity_pairs))     # Summed B(..) values for each pixel.
     B_pixel_count = np.zeros((n_bins, n_bins, n_unique_spin_parity_pairs))   # The number of transitions.
-    rho_ExJpi = np.zeros((n_bins, n_unique_spin_parity_pairs))   # (Ex, Jpi) matrix to store level density
-    gSF = np.zeros((n_bins, n_bins, n_unique_spin_parity_pairs))
+    rho_ExJpi = np.zeros((n_bins, n_unique_spin_parity_pairs))  # (Ex, Jpi) matrix to store level density
+    gSF = np.zeros((n_bins, n_bins, n_unique_spin_parity_pairs))    
 
     for transition_idx in range(n_transitions):
         """
@@ -221,27 +234,25 @@ def gamma_strength_function_average(
         if (Ex_initial_or_final[transition_idx] < Ex_min) or (Ex_initial_or_final[transition_idx] >= Ex_max):
             """
             Check if transition is within min max limits, skip if not.
-            NOTE: Is it correct to check the energy of the final state
-            here? Why not the initial?
             """
             continue
 
-        # Get bin index for Eg and Ex (initial). Indices are defined with respect to the lower bin edge.
-        E_gamma_idx = int(np.floor(transitions[transition_idx, 6]/bin_width))
-        Ex_final_idx = int(np.floor(Ex_initial_or_final[transition_idx]/bin_width))
+        # Get bin index for E_gamma and Ex. Indices are defined with respect to the lower bin edge.
+        E_gamma_idx = int(transitions[transition_idx, 6]/bin_width)
+        Ex_initial_or_final_idx = int(Ex_initial_or_final[transition_idx]/bin_width)
 
-        # Read initial spin and parity of level: NOTE: I think the name / index is wrong. Or do I...?
-        spin_initial = int(transitions[transition_idx, 0])
-        parity_initial = int(transitions[transition_idx, 1])
-        try:
-            """
-            Get index for current [spin_initial, parity_initial]
-            combination in spin_parity_list.
-            """
-            spin_parity_idx = spin_parity_list.index([spin_initial, parity_initial])
-        except ValueError:
-            print("Transition skipped due to lack of spin_parity_list.")
-            continue
+        """
+        transitions : np.ndarray
+            OLD:
+            Mx8 array containing [2*spin_final, parity_initial, Ex_final,
+            2*spin_initial, parity_initial, Ex_initial, E_gamma, B(.., i->f)]
+            NEW:
+            [2*spin_initial, parity_initial, Ex_initial, 2*spin_final,
+            parity_final, Ex_final, E_gamma, B(.., i->f), B(.., f<-i)]
+        """
+        spin_initial = int(transitions[transition_idx, spin_initial_or_final_idx])
+        parity_initial = int(transitions[transition_idx, parity_initial_or_final_idx])
+        spin_parity_idx = spin_parity_list.index([spin_initial, parity_initial])
 
         try:
             """
@@ -249,9 +260,9 @@ def gamma_strength_function_average(
             respectively. NOTE: Hope to remove this try-except by
             implementing suitable input checks to this function.
             """
-            B_pixel_sum[Ex_final_idx, E_gamma_idx, spin_parity_idx] += \
+            B_pixel_sum[Ex_initial_or_final_idx, E_gamma_idx, spin_parity_idx] += \
                 transitions[transition_idx, 7]
-            B_pixel_count[Ex_final_idx, E_gamma_idx, spin_parity_idx] += 1
+            B_pixel_count[Ex_initial_or_final_idx, E_gamma_idx, spin_parity_idx] += 1
         except IndexError as err:
             """
             NOTE: This error usually occurs because Ex_max is set to
@@ -266,39 +277,36 @@ def gamma_strength_function_average(
             is larger.
             """
             msg = f"{err.__str__()}\n"
-            msg += f"{Ex_final_idx=}, {E_gamma_idx=}, {spin_parity_idx=}, {transition_idx=}\n"
+            msg += f"{Ex_initial_or_final_idx=}, {E_gamma_idx=}, {spin_parity_idx=}, {transition_idx=}\n"
             msg += f"{B_pixel_sum.shape=}\n"
             msg += f"{transitions.shape=}\n"
             msg += f"{Ex_max=}\n"
-            msg += f"2*spin_final: {transitions[transition_idx, 0]}\n"
+            msg += f"2*spin_final: {transitions[transition_idx, 3]}\n"
             msg += f"parity_initial: {transitions[transition_idx, 1]}\n"
-            msg += f"Ex_final: {transitions[transition_idx, 2]}\n"
-            msg += f"2*spin_initial: {transitions[transition_idx, 3]}\n"
-            msg += f"parity_initial: {transitions[transition_idx, 4]}\n"
-            msg += f"Ex_initial: {transitions[transition_idx, 5]}\n"
+            msg += f"Ex_final: {transitions[transition_idx, 5]}\n"
+            msg += f"2*spin_initial: {transitions[transition_idx, 0]}\n"
+            msg += f"parity_initial: {transitions[transition_idx, 1]}\n"
+            msg += f"Ex_initial: {transitions[transition_idx, 2]}\n"
             msg += f"E_gamma: {transitions[transition_idx, 6]}\n"
             msg += f"B(.., i->f): {transitions[transition_idx, 7]}\n"
+            msg += f"B(.., f<-i): {transitions[transition_idx, 8]}\n"
             raise Exception(msg) from err
 
     for levels_idx in range(n_levels):
         """
-        Count number of levels for each (Ex, J, parity_initial) pixel.
+        Calculate the level density for each (Ex, spin_parity) pixel.
         """
-
-        if Ex[levels_idx] > Ex_max:
+        if Ex[levels_idx] >= Ex_max:
             """
-            Skip if level is outside range.
+            Skip if level is outside range. Only upper limit since
+            decays to states below the lower limit are allowed.
             """
             continue
 
-        Ex_idx = int(np.floor(Ex[levels_idx]/bin_width))
+        Ex_idx = int(Ex[levels_idx]/bin_width)
 
-        try:
-            spin_parity_idx = \
-                spin_parity_list.index([spins[levels_idx], parities[levels_idx]])
-        except ValueError:
-            print("Transition skipped due to lack of spin_parity_list.")
-            continue
+        spin_parity_idx = \
+            spin_parity_list.index([spins[levels_idx], parities[levels_idx]])
         
         rho_ExJpi[Ex_idx, spin_parity_idx] += 1
 
@@ -306,38 +314,45 @@ def gamma_strength_function_average(
 
     for spin_parity_idx in range(n_unique_spin_parity_pairs):
         """
-        Calculate gamma strength functions for each [Ex, spin,
-        parity_initial] individually using the partial level density for
-        each [spin, parity_initial].
+        Calculate gamma strength functions for each [Ex, E_gamma,
+        spin_parity] individually using the partial level density for
+        each [Ex, spin_parity].
         """
         for Ex_idx in range(n_bins):
             gSF[Ex_idx, :, spin_parity_idx] = \
                 prefactor*rho_ExJpi[Ex_idx, spin_parity_idx]*div0(
-                B_pixel_sum[Ex_idx, :, spin_parity_idx],
-                B_pixel_count[Ex_idx, :, spin_parity_idx]
+                    numerator = B_pixel_sum[Ex_idx, :, spin_parity_idx],
+                    denominator = B_pixel_count[Ex_idx, :, spin_parity_idx]
                 )
 
     # Return the average gSF(Eg) over all (Ex,J,parity_initial)
 
     # return gSF[Ex_min_idx:Ex_max_idx+1,:,:].mean(axis=(0,2))
     # Update 20171009: Took proper care to only average over the non-zero f(Eg,Ex,J,parity_initial) pixels:
-    gSF_currentExrange = gSF[Ex_min_idx:Ex_max_idx + 1, :, :]
+    gSF_currentExrange = gSF[Ex_min_idx:Ex_max_idx + 1, :, :]   # NOTE: Probably not necessary to set an upper limit here due to the input adjustment of Ex_max.
     gSF_ExJpiavg = div0(
-        gSF_currentExrange.sum(axis = (0, 2)),
-        (gSF_currentExrange != 0).sum(axis = (0, 2))
+        numerator = gSF_currentExrange.sum(axis = (0, 2)),
+        denominator = (gSF_currentExrange != 0).sum(axis = (0, 2))  # NOTE: Here is where I left off! Why != 0?
     )
 
-    bins = np.linspace(0, Ex_max, n_bins + 1)   # Not used in this function, only returned.
+    bins = np.linspace(0, Ex_max, n_bins + 1)
     bins = (bins[:-1] + bins[1:])/2   # Middle point of the bins.
     bins = bins[:len(gSF_ExJpiavg)]
 
+    if plot:
+        fig, ax = plt.subplots()
+        ax.plot(bins, gSF_ExJpiavg)
+        ax.set_xlabel("bins")
+        ax.set_ylabel("gsf")
+        plt.show()
+
     return bins, gSF_ExJpiavg
 
 def level_plot(
     levels: np.ndarray,
     max_spin_states: int = 1_000,
     filter_spins: Union[None, list] = None,
-    ax: plt.Axes = None
+    ax: Union[None, plt.Axes] = None
     ):
     """
     Generate a level plot for a single isotope. Spin on the x axis,
@@ -345,19 +360,19 @@ def level_plot(
 
     Parameters
     ----------
-    levels:
+    levels : np.ndarray
         NxM array of [[energy, spin, parity], ...]. This is the instance
         attribute 'levels' of ReadKshellOutput.
     
-    max_spin_states:
+    max_spin_states : int
         The maximum amount of states to plot for each spin. Default set
         to a large number to indicate ≈ no limit.
 
-    filter_spins:
+    filter_spins : Union[None, list]
         Which spins to include in the plot. If None, all spins are
         plotted.
 
-    ax:
+    ax : Union[None, plt.Axes]
         matplotlib Axes to plot on. If None, plt.Figure and plt.Axes is
         generated in this function.
     """
@@ -430,21 +445,21 @@ def level_density(
 
     Parameters
     ----------
-    energy_levels:
+    energy_levels : Union[np.ndarray, list]
         1D array of energy levels.
 
-    bin_size:
+    bin_size : Union[int, float]
         Energy interval of which to calculate the density.
 
-    plot:
+    plot : bool
         For toggling plotting on / off.
 
     Returns
     -------
-    bins:
+    bins : np.ndarray
         The corresponding bins (x value for plotting).
 
-    density:
+    density : np.ndarray
         The level density.
     """
     if isinstance(energy_levels, list):
